from influxdb import InfluxDBClient


class InfluxClient:
    def __init__(self, server, port, database):
        self.server = server
        self.port = port
        self.database = database

        self.client = InfluxDBClient(
            host=self.server,
            port=self.port,
            database=self.database)

    def write_points(self, body):
        self.client.write_points(body)




import logging
import os

import paho.mqtt.client as mqtt

class MqttClient:
    def __init__(self, server, port, username, password, topic, ca_certs, msg_q):
        self.username = username
        self.password = password
        self.topic = topic
        self.ca_certs = ca_certs
        self.server = server
        self.port = port
        self.msg_q = msg_q

        self.logger = logging.getLogger(__name__)

        self.client = mqtt.Client()
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message

    def __del__(self):
        self.client.loop_stop()

    def connect(self):
        if self.username or self.password:
            self.client.username_pw_set(self.username, password=self.password)
        if self.ca_certs != '':
            self.logger("Enabling TLS using '%s'", self.ca_certs)
            self.client.tls_set(ca_certs=self.ca_certs)
        self.client.connect(self.server, self.port, 60)

    def run(self):
        # Blocking call that processes network traffic, dispatches callbacks and
        # handles reconnecting.
        # Other loop*() functions are available that give a threaded interface and a
        # manual interface.
        self.client.loop_start()

    # The callback for when the client receives a CONNACK response from the server.
    def on_connect(self, client, userdata, flags, rc):
        self.logger.info("Connected with result code " + str(rc))

        # Subscribing in on_connect() means that if we lose the connection and
        # reconnect then subscriptions will be renewed.
        client.subscribe(self.topic)

    # The callback for when a PUBLISH message is received from the server.
    def on_message(self, client, userdata, msg):
        self.msg_q.put(msg)



from distutils.core import setup
from glob import glob  # Needed for the examples below only

import setuptools

from energy_usage import VERSION

setup(
    name="energy-usage",
    version=VERSION,
    description="Records realtime energy usage from a Glowmarkt MQTT feed into influxdb",
    author="Ben Roberts",
    author_email="me@benroberts.net",
    url="https://gitlab.sihnon.net/ben/energy-usage",

    # Distribute all python packages within this repository except tests
    packages=setuptools.find_packages(exclude=['tests*']),

    # If necessary, include non-python files stored in your package
    package_data={
        'energy_usage': ['config_default.yaml'],
    },

    # Because the package contains non-python files, it must be installed
    # in uncompressed format, not as a zip-format .egg file
    zip_safe=False,

    # Executable scripts
    # Wrappers will be automatically generated using the correct python shebang
    # to run the methods indicated
    entry_points={
        'console_scripts': [
            'energy-usage=energy_usage.main:main',
        ]
    },

    # Extra content like lib directories, config files from outside of the package
    data_files=[
        ('/etc/energy-usage/config.yaml.example', ['config.yaml.example']),
    ],

    install_requires=[
        'confuse~=1.3',
        'influxdb~=5.3',
        'paho-mqtt~=1.5',
    ]
)



import argparse
import logging
import os
import queue
import sys

import confuse

from energy_usage.influx_client import InfluxClient
from energy_usage.mqtt_client import MqttClient
from energy_usage.sep import parse_sep, usage_to_datapoints


def parse_args(arg_str=None):
    if arg_str is None:
        arg_str = sys.argv[1:]

    parser = argparse.ArgumentParser('energy-usage')

    parser.add_argument(
        '-d', '--debug', dest='debug', action='store_true',
        help='Enable debug output')
    parser.add_argument(
        '-n', '--noop', dest='noop', action='store_true',
        help="Don't make any modifications, just show what would be done")

    args = parser.parse_args(arg_str)

    return parser.prog, args


def load_config(args):
    config = confuse.Configuration('energy-usage', 'energy_usage')
    config.set_args(args)

    validate_config(config)

    return config


def validate_config(config):
    required = [
        (config['mqtt']['username'], str),
        (config['mqtt']['password'], str),
        (config['mqtt']['topic'], str)
    ]

    for item in required:
        key, datatype = item
        key.get(datatype)


def setup_logging(prog, config):
    logger = logging.getLogger()

    if config['debug'].get(bool):
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    handler = logging.StreamHandler(sys.stdout)
    formatter = logging.Formatter('{0}[{1}]: %(levelname)s %(message)s'.format(prog, str(os.getpid())))
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger


def main():
    prog, args = parse_args()

    config = load_config(args)

    logger = setup_logging(prog, config)

    msg_q = queue.Queue()

    mqtt_client = MqttClient(
        server=config['mqtt']['server'].get(str),
        port=config['mqtt']['port'].get(int),
        username=config['mqtt']['username'].get(str),
        password=config['mqtt']['password'].get(str),
        topic=config['mqtt']['topic'].get(str),
        ca_certs=config['ca_certs'].get(str),
        msg_q=msg_q)
    mqtt_client.connect()
    mqtt_client.run()

    influx_client = InfluxClient(
        server=config['influx']['server'].get(str),
        port=config['influx']['port'].get(int),
        database='default',
    )

    while True:
        try:
            msg = msg_q.get()
            logger.debug(msg.topic + " " + str(msg.payload))
            usage = parse_sep(msg.topic, msg.payload)
            if usage:
                usage_datapoints = usage_to_datapoints(usage)
                logger.debug('Writing metrics to influx: %s', usage_datapoints)
                if not config["noop"].get(bool):
                    influx_client.write_points(usage_datapoints)
        except KeyboardInterrupt:
            break


if __name__ == '__main__':
    main()



""" SEP protocol parser
"""
import datetime
import json
import logging


def parse_sep(topic, payload_str):
    """
    
    Credit to https://gist.github.com/ndfred/b373eeafc4f5b0870c1b8857041289a9

    :param topic: str Topic SEP payload was received from
    :param payload_str: str SEP payload message as a string
    :return:
    """
    logger = logging.getLogger(__name__)

    def _get_metric(path, type=int):
        value = payload
        try:
            for hop in path:
                value = value[hop]
            if type is int and type(value) == str:
                return int(value, 16)
            else:
                return type(value)
        except KeyError:
            logger.debug("Ignoring payload missing metric: %s", ".".join(path))
            return None

    payload = json.loads(payload_str)

    if 'gmtime' in payload:
        timestamp = datetime.datetime.fromtimestamp(payload["gmtime"], tz=datetime.timezone.utc)
    else:
        timestamp = datetime.datetime.utcnow()

    if 'gid' in payload:
        device_gid = _get_metric(["gid"], str)
    else:
        device_gid = None

    data = {
        'tags': {
            'topic': topic,
            'gid': device_gid,
        },
    }

    try:
        if 'elecMtr' in payload:
            assert(_get_metric(["elecMtr", "0702", "03", "00"]) == 0)  # kWh

            electricity_consumption = _get_metric(["elecMtr", "0702", "04", "00"])
            electricity_daily_consumption = _get_metric(["elecMtr", "0702", "04", "01"])
            electricity_weekly_consumption = _get_metric(["elecMtr", "0702", "04", "30"])
            electricity_monthly_consumption = _get_metric(["elecMtr", "0702", "04", "40"])
            electricity_multiplier = _get_metric(["elecMtr", "0702", "03", "01"])
            electricity_divisor = _get_metric(["elecMtr", "0702", "03", "02"])
            electricity_meter = _get_metric(["elecMtr", "0702", "00", "00"])
            electricity_mpan = _get_metric(["elecMtr", "0702", "03", "07"], str)

            data.update({'electricity': {
                'tags': {
                    'mpan': electricity_mpan,
                },
                'metrics': {
                    'draw': electricity_consumption * electricity_multiplier / electricity_divisor,
                    'consumption_daily': electricity_daily_consumption * electricity_multiplier / electricity_divisor,
                    'consumption_weekly': electricity_weekly_consumption * electricity_multiplier / electricity_divisor,
                    'consumption_monthly': electricity_monthly_consumption * electricity_multiplier / electricity_divisor,
                    'meter_reading': electricity_meter * electricity_multiplier / electricity_divisor,
                },
            }})
        elif 'electricitymeter' in payload:
            assert(_get_metric(["electricitymeter", "energy", "import", "mpan"]))
            timestamp = datetime.datetime.strptime(payload["electricitymeter"]["timestamp"], '%Y-%m-%dT%H:%M:%SZ')
            data.update({'electricity': {
                'tags': {
                    'mpan': _get_metric(["electricitymeter", "energy", "import", "mpan"]),
                },
                'metrics': {
                    'draw':                _get_metric(["electricitymeter", "power", "value"], float),
                    'consumption_daily':   _get_metric(["electricitymeter", "energy", "import", "day"], float),
                    'consumption_weekly':  _get_metric(["electricitymeter", "energy", "import", "week"], float),
                    'consumption_monthly': _get_metric(["electricitymeter", "energy", "import", "month"], float),
                    'meter_reading':       _get_metric(["electricitymeter", "energy", "import", "cumulative"], float),
                    'unit_rate':           _get_metric(["electricitymeter", "energy", "import", "price", "unitrate"], float),
                    'standing_charge':     _get_metric(["electricitymeter", "energy", "import", "price", "standingcharge"], float),
                },
            }})
    except AssertionError:
        pass

    try:
        if 'gasMtr' in payload:
            assert(_get_metric(["gasMtr", "0702", "03", "01"]) == 1)  # m3
            assert(_get_metric(["gasMtr", "0702", "03", "12"]) == 0)  # kWh
        
            gas_daily_consumption = _get_metric(["gasMtr", "0702", "0C", "01"])
            gas_weekly_consumption = _get_metric(["gasMtr", "0702", "0C", "30"])
            gas_monthly_consumption = _get_metric(["gasMtr", "0702", "0C", "40"])
            gas_multiplier = _get_metric(["gasMtr", "0702", "03", "01"])
            gas_divisor = _get_metric(["gasMtr", "0702", "03", "02"])
            gas_meter = _get_metric(["gasMtr", "0702", "00", "00"])
            gas_mpan = _get_metric(["gasMtr", "0702", "03", "07"], str)

            data.update({'gas': {
                'tags': {
                    'mpan': gas_mpan,
                },
                'metrics': {
                    'consumption_daily': gas_daily_consumption * gas_multiplier / gas_divisor,
                    'consumption_weekly': gas_weekly_consumption * gas_multiplier / gas_divisor,
                    'consumption_monthly': gas_monthly_consumption * gas_multiplier / gas_divisor,
                    'meter_reading': gas_meter * gas_multiplier / gas_divisor,
                },
            }})
        elif 'gasmeter' in payload:
            assert(_get_metric(["gasmeter", "energy", "import", "mprn"]))
            timestamp = datetime.datetime.strptime(payload["gasmeter"]["timestamp"], '%Y-%m-%dT%H:%M:%SZ')
            data.update({'gas': {
                'tags': {
                    'mpan': _get_metric(["gasmeter", "energy", "import", "mprn"]),
                },
                'metrics': {
                    'meter_reading_kwh':   _get_metric(["gasmeter", "energy", "import", "cumulative"], float),
                    'consumption_daily':   _get_metric(["gasmeter", "energy", "import", "day"], float),
                    'consumption_weekly':  _get_metric(["gasmeter", "energy", "import", "week"], float),
                    'consumption_monthly': _get_metric(["gasmeter", "energy", "import", "month"], float),
                    'meter_reading':       _get_metric(["gasmeter", "energy", "import", "cumulativevol"], float),
                    'unit_rate':           _get_metric(["gasmeter", "energy", "import", "price", "unitrate"], float),
                    'standing_charge':     _get_metric(["gasmeter", "energy", "import", "price", "standingcharge"], float),
                },
            }})
    except AssertionError:
        pass

    if 'electricity' not in data and 'gas' not in data:
        logger.warning("Received a payload without either electricity or gas data")
        return None

    data.update({'timestamp': timestamp})

    return data


def usage_to_datapoints(usage):
    datapoints = []

    for utility in ['electricity', 'gas']:
        if utility in usage:
            datapoints.append({
                "measurement": utility,
                "tags": {**usage['tags'], **usage[utility]['tags']},
                "time": usage['timestamp'].isoformat(),
                "fields": usage[utility]['metrics'],
            })

    return datapoints
